/*
 * Exporting captured dns request/responses from libpcap
 * by Javantea
 */

#include <stdio.h>
#include <pcap.h>
#include <netinet/ip.h>
//#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>

// For reading a saved file.
// pcap_t *pcap_open_offline(const char *fname, char *errbuf)
// Close it down.
// void pcap_close(pcap_t *p)
// Get the next packet.
// const u_char *pcap_next(pcap_t *p, struct pcap_pkthdr *h)

// Setup a filter.
// int pcap_compile(pcap_t *p, struct bpf_program *fp,
//       char *str, int optimize, bpf_u_int32 netmask)
// Set the filter.
// int pcap_setfilter(pcap_t *p, struct bpf_program *fp)
// Destroy the filter.
// void pcap_freecode(struct bpf_program *)

// For writing only.
// pcap_dumper_t *pcap_dump_open(pcap_t *p, const char *fname)


/*
Description:
This program reads a pcap file looking for http content sent from a server and 
writes an output file.
TODO:
Create seperate files for each captured file.
	1) Capture GET request.
	2) Get filename from the GET request that begins the transfer, open the file.
	3) At the end of the HTTP (FIN), close the file.
	4) Keep a linked list of active sessions.
Check for bad packets.
	1) Checksum
	2) Missing packets


 */

int main(int argc, char **argv)
{
	printf("Exporting captured dns from libpcap\n");
	printf("by Javantea\n");
	char errbuf[PCAP_ERRBUF_SIZE];
	const char *filename;
	if(argc < 2)
	{
		printf("Usage: pcapdump_dns1 file.cap [filter blah yak yak yak]\n");
		return 1;
	}
	filename = argv[1];
	pcap_t *capture = pcap_open_offline(filename, errbuf);
	if(capture == 0)
	{
		printf("Couldn't open file %s.\n", filename);
		perror("pcap_open_offline");
	}
	
	long remote_addr = 0;
	long local_addr  = 0;
	unsigned short remote_port = 0;
	unsigned short local_port  = 0;
	//struct bpf_program filter;
	if(argc >= 6)
	{
		struct hostent *remote = gethostbyname(argv[2]);
		struct hostent *local  = gethostbyname(argv[3]);
		remote_port = atoi(argv[4]);
		local_port  = atoi(argv[5]);
		if(remote->h_length >= 1) {
			memcpy(&remote_addr, remote->h_addr_list[0], 4);
		}
		if(local->h_length >= 1) {
			memcpy(&local_addr, local->h_addr_list[0], 4);
		}
		/*char ftext[] = "(ip.addr eq 72.246.51.81 and ip.addr eq 192.168.0.4) and (tcp.port eq 80 and tcp.port eq 59481)";
		if(pcap_compile(capture, &filter, ftext, 0, 0) == -1)
		{
			printf("Couldn't parse filter %s: %s\n", ftext, pcap_geterr(capture));
		}
		if(pcap_setfilter(capture, &filter) == -1)
		{
			printf("Couldn't parse filter %s: %s\n", ftext, pcap_geterr(capture));
		}*/
	}
	else
	{
		/*char ftext[] = "";
		if(pcap_compile(capture, &filter, ftext, 0, 0) == -1)
		{
			printf("Couldn't parse filter %s: %s\n", ftext, pcap_geterr(capture));
		}*/
	}
	
	struct pcap_pkthdr header;
	const u_char *packet;
	int i = 0, j = 0;
	FILE *fp = fopen("data_out1.txt", "wb");
	if(fp == 0)
	{
		printf("Unable to open output file: data_out1.txt\n");
		return 1;
	}
	do
	{
		packet = pcap_next(capture, &header);
		if(packet == 0) break;
		j++;
		
		// printf("capture length = %i, packet length = %i\n", header.caplen, header.len);
		//struct ether_arp;
		//printf("dest = %08x, Type = %04hx\n", *(unsigned long *)(packet), *(unsigned short *)(packet+12));
		unsigned short type = ntohs(*(unsigned short *)(packet+12));
		// printf("dest = %04hx%02hhx, Type = %04hx\n", ntohs(*(unsigned short *)(packet)), *(unsigned char *)(packet+2), type);
		if(type != 0x0800) continue;
			
		// Handle IP
		struct iphdr *ip_hdr = (struct iphdr *)(packet+14);
		int ip_header_len = ip_hdr->ihl * 4;
		/*printf("ip {ihl=%i, v=%i, len=%i, proto=%02hhx, src=%08x dst=%08x}\n", 
			ip_header_len, ip_hdr->version, 
			ip_hdr->tot_len, ip_hdr->protocol, 
			ip_hdr->saddr, ip_hdr->daddr);*/
		// UDP
		if(ip_hdr->protocol != 0x11) continue;
		
		// Handle udp
		struct udphdr *udp_hdr = (struct udphdr *)(packet+14+ip_header_len);
		printf("udp {ihl=%i, sport=%i, dport=%i}\n", ip_header_len, udp_hdr->source, udp_hdr->dest);
		if(ntohs(udp_hdr->source) != 53 && ntohs(udp_hdr->dest) != 53) continue;
		
		int udp_header_len = 8;
		int data_len = udp_hdr->len - udp_header_len;
		
		if(remote_addr != 0)
		{
			if(remote_addr != ip_hdr->saddr && remote_addr != ip_hdr->daddr) continue;
			if(local_addr  != ip_hdr->saddr && local_addr  != ip_hdr->daddr) continue;
			if(remote_port != ntohs(udp_hdr->source) && remote_port != ntohs(udp_hdr->dest)) continue;
			if(local_port  != ntohs(udp_hdr->source) && local_port  != ntohs(udp_hdr->dest)) continue;
			// FIXME: Cover the unlikely possibility of flipped ports for address.
		}
		
		// Handle DNS
		printf("udp {j=%i, src=%08x dst=%08x, sport=%i dport=%i hlen=%u, dlen=%i, ulen=%u}\n", 
		       j, ip_hdr->saddr, ip_hdr->daddr,
			   ntohs(udp_hdr->source), ntohs(udp_hdr->dest), 
			   udp_header_len, data_len, ntohs(udp_hdr->len));
		char *data = (char *)(packet+14+ip_header_len+udp_header_len);
		// The format of dns is: len of section, section[len], len[section]...
		// The last one is zero of course.
		// This limits the longest domain and subdomain to be 256.
		// But that's pretty reasonable considering 256 chars ==
		/*
thisisaveryverylongname-dontyouthinkitisaverylongname-ibetitwouldwinsomesortofre
cord-ifyouputitintoalongnamecontest-yesindeed-itisverylong-nowimjustgoingtocopy-
thisisaveryverylongname-dontyouthinkitisaverylongname-ibetitwouldwinsomesortofre
cordifyouputitin
		*/
		if(data_len <= 0) continue;
		int p = 0;
		int j = data[12]+1;
		// TODO: test with very bad data (broken, short packets).
		// Easy buffer overlow if not well-defined, non-exploitable.
		while(j > 1 && p < data_len)
		{
			data[12+p] = '.';
			p += j;
			j = data[12+p]+1;
		}
		printf("dns %s\n", data+12);
		// TODO: Replies are as useful as the queries.
		
		//if(i++ > 2000) break;
	} while(packet != 0);
	
	//pcap_freecode(&filter);
	
	fclose(fp);
	
	pcap_close(capture);
	return 0;
}


