/*
 * Exporting a captured HTTP File from libpcap
 * by Javantea
 */

#include <stdio.h>
#include <pcap.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>

// For reading a saved file.
// pcap_t *pcap_open_offline(const char *fname, char *errbuf)
// Close it down.
// void pcap_close(pcap_t *p)
// Get the next packet.
// const u_char *pcap_next(pcap_t *p, struct pcap_pkthdr *h)

// Setup a filter.
// int pcap_compile(pcap_t *p, struct bpf_program *fp,
//       char *str, int optimize, bpf_u_int32 netmask)
// Set the filter.
// int pcap_setfilter(pcap_t *p, struct bpf_program *fp)
// Destroy the filter.
// void pcap_freecode(struct bpf_program *)

// For writing only.
// pcap_dumper_t *pcap_dump_open(pcap_t *p, const char *fname)


/*
Description:
This program reads a pcap file looking for http content sent from a server and 
writes an output file.
TODO:
Create seperate files for each captured file.
	1) Capture GET request.
	2) Get filename from the GET request that begins the transfer, open the file.
	3) At the end of the HTTP (FIN), close the file.
	4) Keep a linked list of active sessions.
Check for bad packets.
	1) Checksum
	2) Missing packets


 */

int main(int argc, char **argv)
{
	printf("Exporting a captured HTTP File from libpcap\n");
	printf("by Javantea\n");
	char errbuf[PCAP_ERRBUF_SIZE];
	const char *filename;
	if(argc < 2)
	{
		printf("Usage: pcapdump_http1 file.cap [filter blah yak yak yak]\n");
		return 1;
	}
	filename = argv[1];
	pcap_t *capture = pcap_open_offline(filename, errbuf);
	if(capture == 0)
	{
		printf("Couldn't open file %s.\n", filename);
		perror("pcap_open_offline");
	}
	
	long remote_addr = 0;
	long local_addr  = 0;
	unsigned short remote_port = 0;
	unsigned short local_port  = 0;
	//struct bpf_program filter;
	if(argc >= 6)
	{
		struct hostent *remote = gethostbyname(argv[2]);
		struct hostent *local  = gethostbyname(argv[3]);
		remote_port = atoi(argv[4]);
		local_port  = atoi(argv[5]);
		if(remote->h_length >= 1) {
			memcpy(&remote_addr, remote->h_addr_list[0], 4);
		}
		if(local->h_length >= 1) {
			memcpy(&local_addr, local->h_addr_list[0], 4);
		}
		/*char ftext[] = "(ip.addr eq 72.246.51.81 and ip.addr eq 192.168.0.4) and (tcp.port eq 80 and tcp.port eq 59481)";
		if(pcap_compile(capture, &filter, ftext, 0, 0) == -1)
		{
			printf("Couldn't parse filter %s: %s\n", ftext, pcap_geterr(capture));
		}
		if(pcap_setfilter(capture, &filter) == -1)
		{
			printf("Couldn't parse filter %s: %s\n", ftext, pcap_geterr(capture));
		}*/
	}
	else
	{
		/*char ftext[] = "";
		if(pcap_compile(capture, &filter, ftext, 0, 0) == -1)
		{
			printf("Couldn't parse filter %s: %s\n", ftext, pcap_geterr(capture));
		}*/
	}
	
	struct pcap_pkthdr header;
	const u_char *packet;
	int i = 0, j = 0;
	FILE *fp = fopen("data_out1.bin", "wb");
	if(fp == 0)
	{
		printf("Unable to open output file: data_out1.bin\n");
		return 1;
	}
	do
	{
		packet = pcap_next(capture, &header);
		if(packet == 0) break;
		j++;
		
		// printf("capture length = %i, packet length = %i\n", header.caplen, header.len);
		//struct ether_arp;
		//printf("dest = %08x, Type = %04hx\n", *(unsigned long *)(packet), *(unsigned short *)(packet+12));
		unsigned short type = ntohs(*(unsigned short *)(packet+12));
		// printf("dest = %04hx%02hhx, Type = %04hx\n", ntohs(*(unsigned short *)(packet)), *(unsigned char *)(packet+2), type);
		if(type != 0x0800) continue;
			
		// Handle IP
		struct iphdr *ip_hdr = (struct iphdr *)(packet+14);
		int ip_header_len = ip_hdr->ihl * 4;
		/*printf("ip {ihl=%i, v=%i, len=%i, proto=%02hhx, src=%08x dst=%08x}\n", 
			ip_header_len, ip_hdr->version, 
			ip_hdr->tot_len, ip_hdr->protocol, 
			ip_hdr->saddr, ip_hdr->daddr);*/
		if(ip_hdr->protocol != 0x06) continue;
		
		// Handle TCP
		struct tcphdr *tcp_hdr = (struct tcphdr *)(packet+14+ip_header_len);
		unsigned char flags = (tcp_hdr->fin << 5) | (tcp_hdr->syn << 4) | (tcp_hdr->rst << 3) | (tcp_hdr->psh << 2) | (tcp_hdr->ack << 1) | (tcp_hdr->urg);
		//printf("tcp {ihl=%i, sport=%i, dport=%i}\n", ip_header_len, tcp_hdr->source, tcp_hdr->dest);
		if(ntohs(tcp_hdr->source) != 80 && ntohs(tcp_hdr->dest) != 80) continue;
		
		int tcp_header_len = tcp_hdr->doff * 4;
		int data_len = ntohs(ip_hdr->tot_len) - ip_header_len - tcp_header_len;
		
		if(remote_addr != 0)
		{
			if(remote_addr != ip_hdr->saddr && remote_addr != ip_hdr->daddr) continue;
			if(local_addr  != ip_hdr->saddr && local_addr  != ip_hdr->daddr) continue;
			if(remote_port != ntohs(tcp_hdr->source) && remote_port != ntohs(tcp_hdr->dest)) continue;
			if(local_port  != ntohs(tcp_hdr->source) && local_port  != ntohs(tcp_hdr->dest)) continue;
			// FIXME: Cover the unlikely possibility of flipped ports for address.
		}
		
		// Handle HTTP
		printf("tcp {j=%i, src=%08x dst=%08x, sport=%i dport=%i, seq=%u, ack=%u hlen=%u, ctrl=%02hhx, dlen=%i, win=%u}\n", 
		       j, ip_hdr->saddr, ip_hdr->daddr,
		       ntohs(tcp_hdr->source), ntohs(tcp_hdr->dest), 
			ntohl(tcp_hdr->seq), ntohl(tcp_hdr->ack_seq), 
			tcp_header_len, flags, data_len, ntohs(tcp_hdr->window));
		char *data = (char *)(packet+14+ip_header_len+tcp_header_len);
		// Bad idea --> || (!tcp_hdr->psh || !tcp_hdr->ack)
		if(data_len <= 0) continue;
		fwrite(data, data_len, 1, fp);
		
		//if(i++ > 2000) break;
	} while(packet != 0);
	
	//pcap_freecode(&filter);
	
	fclose(fp);
	
	pcap_close(capture);
	return 0;
}


